/*
 * Copyright (c) 2016 Hanspeter Portner (dev@open-music-kontrollers.ch)
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the Artistic License 2.0 as published by
 * The Perl Foundation.
 *
 * This source is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * Artistic License 2.0 for more details.
 *
 * You should have received a copy of the Artistic License 2.0
 * along the source as a COPYING file. If not, obtain it from
 * http://www.perlfoundation.org/artistic_license_2_0.
 */

#ifndef _TRACKER_API_ATOM_H
#define _TRACKER_API_ATOM_H

#include <tracker.h>

#include <lauxlib.h>

#define DRIVER_HASH_MAX 15

typedef struct _latom_driver_t latom_driver_t;
typedef struct _lseq_t lseq_t;
typedef struct _latom_t latom_t;
typedef struct _lobj_t lobj_t;
typedef struct _ltuple_t ltuple_t;
typedef struct _lvec_t lvec_t;
typedef struct _lroot_t lroot_t;
typedef struct _latom_driver_hash_t latom_driver_hash_t;

typedef int (*latom_driver_function_t)(lua_State *L, latom_t *latom);
typedef int (*latom_driver_function_indexk_t)(lua_State *L, latom_t *latom, const char *key);

struct _latom_driver_t {
	latom_driver_function_t __indexi;
	latom_driver_function_indexk_t __indexk;
	latom_driver_function_t __len;
	latom_driver_function_t __tostring;
	latom_driver_function_t __call;

	latom_driver_function_t value;
	//int unpack;
	//int foreach;
	lua_CFunction unpack;
	lua_CFunction foreach;
};

struct _latom_t {
	const LV2_Atom *atom;

	union {
		const void *raw;

		const int32_t *i32;
		const int64_t *i64;
		const float *f32;
		const double *f64;
		const uint32_t *u32;
		const char *str;
		const LV2_Atom_Literal_Body *lit;

		const LV2_Atom_Sequence_Body *seq;
		const LV2_Atom_Object_Body *obj;
		const LV2_Atom *tuple;
		const LV2_Atom_Vector_Body *vec;
	} body;

	union {
		struct _lseq_t {
			const LV2_Atom_Event *ev;
		} seq;

		struct _lobj_t {
			const LV2_Atom_Property_Body *prop;
		} obj;

		struct _ltuple_t {
			int pos;
			const LV2_Atom *item;
		} tuple;

		struct _lvec_t {
			int count;
			int pos;
		} vec;
	} iter;

	LV2_Atom payload [0];
};

struct _latom_driver_hash_t {
	LV2_URID type;
	const latom_driver_t *driver;
};

struct _lroot_t {
	struct {
		LV2_URID atom_beat_time;
	} uris;

	LV2_Atom_Forge forge;

	latom_driver_hash_t atom_driver_hash [DRIVER_HASH_MAX];
};

extern const luaL_Reg latom_mt [];

latom_t *
latom_newuserdata(lua_State *L, const LV2_Atom *atom);

void
lroot_init(lroot_t *lroot, LV2_URID_Map *map);


#endif
