/*
 * Copyright (c) 2016 Hanspeter Portner (dev@open-music-kontrollers.ch)
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the Artistic License 2.0 as published by
 * The Perl Foundation.
 *
 * This source is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * Artistic License 2.0 for more details.
 *
 * You should have received a copy of the Artistic License 2.0
 * along the source as a COPYING file. If not, obtain it from
 * http://www.perlfoundation.org/artistic_license_2_0.
 */

#ifndef _TRACKER_API_FORGE_H
#define _TRACKER_API_FORGE_H

#include <tracker.h>

#include <lauxlib.h>

#include <osc.lv2/osc.h>

typedef struct _patch_t patch_t;
typedef struct _lbase_t lbase_t;
typedef struct _lforge_t lforge_t;

struct _patch_t {
	LV2_URID self;

	LV2_URID get;
	LV2_URID set;
	LV2_URID put;
	LV2_URID patch;
	LV2_URID body;
	LV2_URID subject;
	LV2_URID property;
	LV2_URID value;
	LV2_URID add;
	LV2_URID remove;
	LV2_URID wildcard;
	LV2_URID writable;
	LV2_URID readable;
	LV2_URID destination;
	LV2_URID sequence;
	LV2_URID error;
	LV2_URID ack;
};

struct _lbase_t {
	struct {
		LV2_URID midi_event;
		patch_t patch;

		LV2_URID atom_beat_time;
		LV2_URID atom_frame_time;
	} uris;

	LV2_Atom_Forge forge;

	LV2_OSC_URID osc_urid;
	LV2_OSC_Schedule *osc_sched;

	LV2_URID_Map *map;
};

struct _lforge_t {
	LV2_Atom_Forge *forge;

	int depth;

	union {
		int64_t frames;	// Time in audio frames
		double  beats; // Time in beats
	} last;

	LV2_Atom_Forge_Frame frame [2];
};

void
lbase_init(lbase_t *lbase, const char *uri, LV2_URID_Map *map);

lforge_t *
lforge_newuserdata(lua_State *L);

extern const luaL_Reg lforge_mt [];

#endif
